# # SPDX-FileCopyrightText: 2025 Maxime Letellier <maxime.eliot.letellier@gmail.com>
#
# # SPDX-License-Identifier: GPL-3.0-or-later

from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from subprocess import run
from typing import Dict, List

from tomlkit import TOMLDocument
from tomlkit.toml_file import TOMLFile

from peeler.utils import restore_file
from peeler.uv_utils import find_uv_bin

LOCK_FILE = "uv.lock"


def _get_lock_path(pyproject_file: Path) -> Path:
    return Path(pyproject_file).parent / LOCK_FILE


@contextmanager
def _get_lock_file(pyproject_file: Path) -> Generator[Path, None, None]:
    uv_bin = find_uv_bin()

    lock_path = _get_lock_path(pyproject_file)

    with restore_file(lock_path, missing_ok=True):
        run(
            [
                uv_bin,
                "--no-config",
                "--directory",
                pyproject_file.parent,
                "--no-python-downloads",
                "lock",
                "--no-build",
            ],
            cwd=pyproject_file.parent,
        )

        yield lock_path


def _get_wheels_urls_from_lock(lock_toml: TOMLDocument) -> Dict[str, List[str]]:
    urls: Dict[str, List[str]] = {}

    if (packages := lock_toml.get("package", None)) is None:
        return {}

    for package in packages:
        if "wheels" not in package:
            continue

        urls[package["name"]] = [wheels["url"] for wheels in package["wheels"]]

    return urls


def get_wheels_url(pyproject_file: Path) -> Dict[str, List[str]]:
    """Return a Dict containing wheels urls from a pyproject.toml dependencies table.

    :param pyproject_file: the pyproject file.
    :return: A Dict with package name as key and a list of package urls as values.
    """

    with _get_lock_file(pyproject_file) as lock_file:
        # open lock file to retrieve wheels url for all platform
        lock_toml = TOMLFile(lock_file).read()

    return _get_wheels_urls_from_lock(lock_toml)



# # SPDX-FileCopyrightText: 2025 Maxime Letellier <maxime.eliot.letellier@gmail.com>
#
# # SPDX-License-Identifier: GPL-3.0-or-later

import os
from pathlib import Path
from subprocess import run
from typing import Dict, List, Tuple

import typer
from click import ClickException
from typer import progressbar
from wheel_filename import parse_wheel_filename

from peeler.uv_utils import find_uv_bin

_VALID_IMPLEMENTATIONS = {"cp", "py"}


def _parse_implementation_and_python_version(python_tag: str) -> Tuple[str, str]:
    return python_tag[:2], python_tag[2:]


def _has_valid_implementation(url: str) -> bool:
    wheel_info = parse_wheel_filename(url)

    return any(
        _parse_implementation_and_python_version(tag)[0].lower()
        in _VALID_IMPLEMENTATIONS
        for tag in wheel_info.python_tags
    )


def _download_from_url(destination_directory: Path, url: str) -> Path:
    wheel_info = parse_wheel_filename(url)
    path = destination_directory / str(wheel_info)

    if path.is_file():
        return path

    platform = wheel_info.platform_tags[0]
    implementation, python_version = _parse_implementation_and_python_version(
        wheel_info.python_tags[0]
    )
    abi = wheel_info.abi_tags[0]

    _destination_directory = os.fspath(destination_directory.resolve())
    uv_bin = find_uv_bin()
    cmd = [
        uv_bin,
        "--isolated",
        "tool",
        "run",
        "--no-config",
        "--no-python-downloads",
        "--no-build",
        "pip",
        "download",
        "-d",
        _destination_directory,
        "--no-deps",
        "--only-binary",
        ":all:",
        "--platform",
        platform,
        "--abi",
        abi,
        "--implementation",
        implementation,
        "--progress-bar",
        "off",
    ]

    if len(python_version) > 1:
        cmd.extend(["--python-version", python_version])

    cmd.append(url)

    result = run(cmd, capture_output=True, text=True)
    stderr = result.stderr

    if not path.is_file():
        msg = f"Error when downloading wheel for package `{wheel_info.project}` for platform `{platform}`"
        raise ClickException(f"{msg}{stderr}")

    return path


def download_wheels(wheels_directory: Path, urls: Dict[str, List[str]]) -> List[Path]:
    """Download the wheels from urls with pip download into wheels_directory.

    :param wheels_directory: The directory to download wheels into
    :param urls: A Dict with package name as key and a list of package urls as values.
    :return: the list of the downloaded wheels path
    """
    wheels_directory.mkdir(parents=True, exist_ok=True)

    wheels_paths: List[Path] = []

    for package_name, package_urls in urls.items():
        # filter out python implementations not supported by blender
        package_urls = list(filter(_has_valid_implementation, package_urls))

        if not package_urls:
            msg = (
                f"No suitable implementation found for {package_name}, not downloading."
            )
            typer.echo(f"Warning: {msg}")
            continue

        with progressbar(package_urls, label=package_name, color=True) as _package_urls:
            for url in _package_urls:
                filename = _download_from_url(wheels_directory, url)

                wheels_paths.append(filename)

    return wheels_paths



# # SPDX-FileCopyrightText: 2025 Maxime Letellier <maxime.eliot.letellier@gmail.com>
#
# # SPDX-License-Identifier: GPL-3.0-or-later

import re
import shutil
from os import PathLike, fspath
from pathlib import Path
from subprocess import run

from click import ClickException
from packaging.version import Version

from peeler import MAX_UV_VERSION, MIN_UV_VERSION

version_regex = r"(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?"


def get_uv_bin_version(uv_bin: PathLike) -> Version | None:
    """Return the uv version.

    :param uv_bin: path to a uv bin
    :return: the version of the provided binary
    """

    uv_bin = fspath(uv_bin)

    result = run([uv_bin, "version"], capture_output=True, text=True, check=True)
    output = result.stdout.strip()
    match = re.search(version_regex, output)

    if not match:
        return None

    return Version(match.group(0))


def find_uv_bin() -> str:
    """Return the path to the uv bin.

    :raises ClickException: if the bin cannot be found.
    """

    try:
        import uv

        uv_bin = uv._find_uv.find_uv_bin()
    except (ModuleNotFoundError, FileNotFoundError):
        uv_bin = shutil.which("uv")

    if uv_bin is None:
        raise ClickException(
            f"""Cannot find uv bin
Install uv `https://astral.sh/blog/uv` or
Install peeler optional dependency uv (eg: pip install peeler[uv])
"""
        )

    return uv_bin


def get_uv_version() -> Version | None:
    """Return uv version."""

    return get_uv_bin_version(Path(find_uv_bin()))


def check_uv_version() -> None:
    """Check the current uv version is between 0.5.17 and current supported max uv version.

    See .max-uv-version or pyproject.toml files.

    :raises ClickException: if uv version cannot be determined or is lower than the minimum version.
    """

    uv_version = get_uv_bin_version(Path(find_uv_bin()))

    try:
        import uv
    except (ModuleNotFoundError, FileNotFoundError):
        from_pip = False
    else:
        from_pip = True

    import peeler

    body = f"To use {peeler.__name__} wheels feature uv version must be between {MIN_UV_VERSION} and {MAX_UV_VERSION}"

    if from_pip:
        update_uv = """Install peeler with a supported uv version:

pip install peeler[uv]"""
    else:
        update_uv = f"""Use peeler with a supported uv version without changing your current uv installation:

uvx peeler[uv] [OPTIONS] COMMAND [ARGS]"""

    if not uv_version:
        header = "Error when checking uv version. Make sur to have installed, visit: https://docs.astral.sh/uv/getting-started/installation/"
        raise ClickException(f"""{header}

{body}

{update_uv}""")

    if uv_version > MAX_UV_VERSION or uv_version < MIN_UV_VERSION:
        import peeler

        header = f"uv version is {uv_version}"

        raise ClickException(f"""{header}
{body}
{update_uv}""")
